pro ng_2ndcalib,wave_array,ADU_1d,d_ADU_1d,source_table, $
                flux_1d,d_flux_1d_adu,d_flux_1d_res, $
                DIAGONAL=diagonal

; S. Baba (20160908)
; This function calibrate raw spectra in ADU not ignoring
; the 2nd-order light contamination.
;
; inputs
;   wave_array: a vector of 1st-order wavelengths
;   ADU_1d: a array of 1D raw spectra ([wavelength, source id])
;   d_ADU_1d: a array of uncertainty of ADU_1d
;   source_table: a structure of imformation of the souces
; outputs
;   flux_1d: a array of 1D flux-calibrated spectra
;   d_flux_1d_adu: a array of uncertainty of flux_1d stems from d_ADU_1d
;   d_flux_1d_res: a array of uncertainty of flux_1d stems from d_response
; option
;   DIAGONAL: If this keyword is set, the flux calibration is executed
;             with only diagonal elements of the response matrix.
;             This means that the 2nd-order response is ignored and that
;             the gross ADU is treated as 1st-order component.
;
  
  ; number of sources
  n_source=n_elements(source_table)
  
  ; define output arrays
  flux_1d       = fltarr(size(ADU_1d,/dimensions))
  d_flux_1d_adu = fltarr(size(ADU_1d,/dimensions))
  d_flux_1d_res = fltarr(size(ADU_1d,/dimensions))
  
  ; read spectral response functions
  set_response,response_lambda_1st,response_1st,d_response_1st
  set_response,response_lambda_2nd,response_2nd,d_response_2nd,/NG_2ND
  
  wave_array_1st = wave_array
  wave_array_2nd = ng_waveconv(wave_array)
  
  for source_id=0,n_source-1 do begin
    ; restore ADU spectra as a vector
      adu_tmp =   ADU_1d[*,source_id]
    d_adu_tmp = d_ADU_1d[*,source_id]
    ; remove bad subscripts
    valid = where(wave_array_1st gt min(response_lambda_1st) and $
                  wave_array_1st lt max(response_lambda_1st) and $
                  wave_array_2nd gt min(response_lambda_2nd) and $
                  wave_array_2nd lt max(response_lambda_2nd) and $
                  finite(adu_tmp) and finite(d_adu_tmp))
      adu =   adu_tmp[valid]
    d_adu = d_adu_tmp[valid]
    lambda_1st = wave_array_1st[valid]
    lambda_2nd = wave_array_2nd[valid]
    ; interpolate responses
    linterp,response_lambda_1st,  response_1st,lambda_1st,  response_at_lambda_1st
    linterp,response_lambda_1st,d_response_1st,lambda_1st,d_response_at_lambda_1st
    linterp,response_lambda_2nd,  response_2nd,lambda_2nd,  response_at_lambda_2nd
    linterp,response_lambda_2nd,d_response_2nd,lambda_2nd,d_response_at_lambda_2nd
    ; make a response matrix and its uncertainty
    ng_make_response_matrix,lambda_1st,              lambda_2nd, $
                            response_at_lambda_1st,  response_at_lambda_2nd, $
                            d_response_at_lambda_1st,d_response_at_lambda_2nd, $
                            response_matrix, $
                            d_response_matrix, $
                            DIAGONAL=diagonal
    ; invert the response matrix and its uncertainty analytically
    ng_make_dullness_matrix,response_matrix,d_response_matrix, $
                            dullness_matrix,d_dullness_matrix, $
                            DIAGONAL=diagonal
    ; transpose row vectors to column vectors
    adu = transpose(adu)
    d_adu = transpose(d_adu)
    ; multiply the invert matrix of the response matrix
    flux = dullness_matrix ## adu
    d_flux_from_d_adu = sqrt(  dullness_matrix^2. ## d_adu^2. )
    d_flux_from_d_res = sqrt(d_dullness_matrix^2. ##   adu^2. )
    ; transpose column vectors to row vectors
    flux = transpose(flux)
    d_flux_from_d_adu = transpose(d_flux_from_d_adu)
    d_flux_from_d_res = transpose(d_flux_from_d_res)
    ; substitute
    flux_1d[valid,source_id]=flux
    d_flux_1d_adu[valid,source_id]=d_flux_from_d_adu
    d_flux_1d_res[valid,source_id]=d_flux_from_d_res
  endfor
  
end